from typing import List
from uuid import uuid4

import pandas as pd
from toolz import curry

from recette.types import Dtype, StepType


@curry
def prep_step_other(
    df: pd.DataFrame,
    cols: List[str],
    threshold: float = 0.05,
    other_cat: str = "other",
    ignore_na: bool = False,
) -> StepType:
    """Collapse categories with proportion below threshold.

    Args:
        df: Pandas DataFrame to process.
        cols: Columns to process.
        threshold: Proportion threshold below which categories are collapsed.
        other_cat: Category used for collapsed categories.
        ignore_na: Ignore NAs when computing proportions.

    Returns:
        Step that collapses all categories other than the ones above proportion
        threshold.

    """
    # Get categories above proportion threshold for each column
    above_threshold_cats = {
        col: df.loc[:, col]
        .value_counts(normalize=True, dropna=ignore_na)
        .loc[lambda srs: srs >= threshold]
        .index
        for col in cols
    }

    def step_other(df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        for col in cols:
            # Add other category if it isn't already present
            if other_cat not in df.loc[:, col].cat.categories:
                df.loc[:, col] = df.loc[:, col].cat.add_categories(other_cat)

            # Find out which rows should be collapsed
            is_other_cat = ~(
                df.loc[:, col].isin(above_threshold_cats[col]) | df.loc[:, col].isna()
            )

            # Collapse categories and remove them
            df.loc[is_other_cat, col] = other_cat
            df.loc[:, col] = df.loc[:, col].cat.remove_unused_categories()

        return df

    return step_other


@curry
def prep_step_dummy(
    df: pd.DataFrame,
    cols: List[str],
    drop_first: bool = True,
    keep_cols: bool = False,
    sep: str = "=",
    dtype: Dtype = "UInt8",
) -> StepType:
    """Convert categorical columns to dummy columns.

    Args:
        df: Pandas DataFrame to process.
        cols: Columns to process.
        drop_first: For k categories, create k - 1 dummy variables instead of k
            by removing the first category.
        keep_cols: Keep processed columns (in addition to dummy columns).
        sep: Separator between column name and category for dummy column names.
        dtype: Dummies data type. Note that if NAs are expected, a nullable data
            type should be used.

    Returns:
        Step that converts categorical columns to dummy columns.

    """
    # Store known categories for each column, dropping the first if appropriate
    cols_cats = [df.loc[:, col].cat.categories[int(drop_first) :] for col in cols]

    def step_dummy(df: pd.DataFrame) -> pd.DataFrame:
        original_df, df = df, df.copy()

        # Convert all unknown categories to temporary ones
        tmp_cats = []
        for col, cats in zip(cols, cols_cats):
            # Get unknown categories
            unknown_cats = df.loc[:, col].cat.categories.difference(cats)

            # Make temporary category
            tmp_cat = str(uuid4())  # Ensure no collapse with any category
            tmp_cats.append(tmp_cat)
            df.loc[:, col] = df.loc[:, col].cat.add_categories(tmp_cat)

            # Convert unknown categories to it
            df.loc[df.loc[:, col].isin(unknown_cats), col] = tmp_cat

            # Remove unknown categories and set the remaining categories order
            df.loc[:, col] = (
                df.loc[:, col]
                .cat.remove_categories(unknown_cats)
                .cat.set_categories(cats.insert(0, tmp_cat))
            )

        dummies_df = pd.get_dummies(df.loc[:, cols], columns=cols, prefix_sep=sep)

        # Remove temporary category dummies
        tmp_dummy_cols = dummies_df.columns.intersection(
            [col + sep + str(tmp_cat) for col, tmp_cat in zip(cols, tmp_cats)]
        )
        dummies_df = dummies_df.drop(columns=tmp_dummy_cols)

        # Propagate NAs
        for col, cats in zip(cols, cols_cats):
            dummy_cols = [col + sep + str(cat) for cat in cats]
            dummies_df.loc[df.loc[:, col].isna(), dummy_cols] = pd.NA

        # Cast dummies to given dtype
        dummies_df = dummies_df.astype(dtype)

        return original_df.drop(columns=[] if keep_cols else cols).join(dummies_df)

    return step_dummy



from typing import Union

import pandas as pd

from recette.types import PrepType, StepType


def combine(*step_or_preps: Union[StepType, PrepType]) -> PrepType:
    """Combine a sequence of steps and preparations into a single preparation.

    Args:
        step_or_preps: Sequence of steps or preparations.

    Returns:
        Preparation applying each given steps or preparation in sequence.

    """

    def prep(df: pd.DataFrame) -> StepType:
        steps = []
        df_or_step = df
        for step_or_prep in iter(step_or_preps):
            # Apply step or prep to DataFrame, returning a DataFrame or a step
            df_or_step = step_or_prep(df)
            if isinstance(df_or_step, pd.DataFrame):
                # Store step in case it's a step
                steps.append(step_or_prep)
                df = df_or_step
            else:
                # Else, apply step to DataFrame and store step
                steps.append(df_or_step)
                df = df_or_step(df)

        def step(df: pd.DataFrame) -> pd.DataFrame:
            # Apply each step in sequence
            for step in steps:
                df = step(df)

            return df

        return step

    return prep



import codecs
import os.path

import setuptools


# Functions used to parse the package version from its top __init__.py following
# the approach in:
# https://packaging.python.org/guides/single-sourcing-package-version/
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as f:
        return f.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="recette",
    version=get_version("recette/__init__.py"),
    author="Alan Pennacchio",
    author_email="alanpennacchio@icloud.com",
    description="Data preprocessing utilities on top of pandas.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pennacchio/recette",
    packages=["recette"],
    python_requires=">=3.6",
    install_requires=["pandas >= 1.0.3", "toolz >= 0.10.0"],
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
